#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <errno.h>

#include "plog-export.h"

/* Decodes PLOG's 100-mas number and returns in turns (1 = 360 degrees). */
static inline double
decode_100mas(int ix)
{
    return (int) ntohl(ix) / 12960000.0;
}

/* Decodes PLOG's 10-ms number and returns in turns (1 = 24 hours). */
static inline double
decode_10ms(int ix)
{
    return (int) ntohl(ix) / 8640000.0;
}

/* Takes care of short read. */
int
readn(int s, void* buff, size_t size)
{
    int n;
    char *p;

    for (p = (char*) buff, n = size; n > 0;)
    {
	int m;

	if ((m = read(s, p, n)) < 0)
	{
	    if (errno == EINTR)
		continue;
	    return -1;
	}
	else if (m == 0)
	{
#if 0
	    return size - n;	/* Returns short count. */
#else
	    return 0;		/* Returns 0 if EOF. */
#endif
	}
	n -= m;
	p += m;
    }
    return size;
}

/* Connects to TCP service. */
int
tcp_client(const char* hostname, const char* servname)
{
    struct hostent* hostent_;
    struct servent* servent_;
    struct sockaddr_in sockaddr_;
    unsigned short port;
    int s;
    const char* proto = "tcp";

    memset(&sockaddr_, 0, sizeof (sockaddr_));
    sockaddr_.sin_family = AF_INET;

    if ((sockaddr_.sin_addr.s_addr = inet_addr(hostname)) != -1)
	/*EMPTY*/;
    else if ((hostent_ = gethostbyname(hostname)) != 0)
	memcpy(&sockaddr_.sin_addr, hostent_->h_addr,
	       sizeof (sockaddr_.sin_addr));
    else
    {
	fprintf(stderr, "gethostbyname: %s (%s)\n",
		hstrerror(h_errno), hostname);
	return -1;
    }

    if ((port = (unsigned short) atoi(servname)) > 0)
	sockaddr_.sin_port = htons(port);
    else if ((servent_ = getservbyname(servname, proto)) != 0)
	sockaddr_.sin_port = servent_->s_port;
    else
    {
	fprintf(stderr, "getservbyname: Unknown service (%s/%s)\n",
		servname, proto);
	return -1;
    }

    if ((s = socket(AF_INET, SOCK_STREAM, PF_UNSPEC)) < 0)
    {
	perror("socket");
	return -1;
    }

    if (connect(s, (struct sockaddr*) &sockaddr_, sizeof (sockaddr_)) == -1)
    {
	perror("connect");
	return -1;
    }

    return s;
}

static void
SIGINT_handler(int signo)
{
    exit(0);
}

int
main()
{
    int s;

    if (signal(SIGINT, SIGINT_handler) == SIG_ERR)
    {
	perror("signal");
	return 1;
    }

    /* Connects to CSO antenna computer's PLOG server (100-Hz version). */
    if ((s = tcp_client("128.171.86.90", "50020")) < 0)
	return 1;

    for (;;)
    {
	/* Update rate is 1-Hz (100 entries per read). */
#define BLOCK_SIZE (100)

	static struct plogent_export ee[BLOCK_SIZE];
	struct plogent_export* pee;

	if (readn(s, ee, sizeof (ee)) < 0)
	    break;

	for (pee = ee; pee != ee + BLOCK_SIZE; ++pee)
	{
	    unsigned int flags = ntohl(pee->flags);

	    /* UTC modified Julian day at midnight */
	    printf("%5d", ntohl(pee->epoch));
	    /* UTC second of day */
	    printf(" %8.2f", (double) ntohl(pee->ticks) / 100);
	    /* Some flag bits */
	    printf(" %c%c%c%c%c",
		   flags & EXPLOG_TRACKING ? 'T' : 'F',
		   flags & EXPLOG_ACQUIRED ? 'T' : 'F',
		   flags & EXPLOG_SCANNING ? 'T' : 'F',
		   flags & EXPLOG_TRANSITED ? 'T' : 'F',
		   flags & EXPLOG_CELESTIAL ? 'T' : 'F');
	    /* LST in hours */
	    printf(" %9.6f", decode_10ms(pee->sidereal_time) * 24);
	    /* Right ascension in hours */
	    printf(" %9.6f", decode_10ms(pee->right_ascension) * 24);
	    /* Declination in degrees */
	    printf(" %+9.5f", decode_100mas(pee->declination) * 360);
	    /* Parallactic angle in degrees */
	    printf(" %+10.5f", decode_100mas(pee->parallactic_angle) * 360);
	    /* X offset in arcseconds */
	    printf(" %+10.1f", decode_100mas(pee->x_offset) * 1296000);
	    /* Y offset in arcseconds */
	    printf(" %+9.1f", decode_100mas(pee->y_offset) * 1296000);
	    /* Azimuth in degrees */
	    printf(" %+10.5f", decode_100mas(pee->azimuth) * 360);
	    /* Elevation in degrees */
	    printf(" %+9.5f", decode_100mas(pee->elevation) * 360);
	    /* Azimuth error in arcseconds */
	    printf(" %+10.1f", decode_100mas(pee->azimuth_error) * 1296000);
	    /* Elevation error in arcseconds */
	    printf(" %+9.1f", decode_100mas(pee->elevation_error) * 1296000);
	    printf("\r\n");
	}
    }

    return 0;
}
